from typing import List, Dict, TYPE_CHECKING

from PyQt5.QtCore import Qt, pyqtSignal, QAbstractListModel, QModelIndex
from PyQt5.QtWidgets import QApplication

from astra_installer.settings.component import Component

if TYPE_CHECKING:
    from gettext import gettext as _

translate = QApplication.translate


class ComponentListModel(QAbstractListModel):
    checked_components_changed = pyqtSignal()

    def __init__(self) -> None:
        super(ComponentListModel, self).__init__()
        self.__components = []  # type: List[Component]
        self.__index_by_id = {}  # type: Dict[str, int]

    @property
    def components(self) -> List[Component]:
        return self.__components

    @components.setter
    def components(self, values: List[Component]) -> None:
        self.beginResetModel()
        self.__components = values
        self.__index_by_id.clear()
        for index, component in enumerate(values):
            self.__index_by_id.update({component.ident: index})
        self.endResetModel()

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self.components)

    def data(self, index: QModelIndex, role: Qt.ItemDataRole = Qt.DisplayRole) -> object:
        if not index.isValid():
            return None
        component = self.components[index.row()]
        if role == Qt.DisplayRole:
            return _(component.name)
        elif role == Qt.ToolTipRole:
            if component.desc:
                return _(component.desc)
        elif role == Qt.CheckStateRole:
            return self.component_check_state_role(component)
        elif role == Qt.UserRole:
            return component
        else:
            return None

    def setData(self, index: QModelIndex, value: object, role: int = Qt.EditRole) -> bool:
        if role == Qt.CheckStateRole:
            component = self.components[index.row()]
            self.set_component_check_state_role(component, value)
            self.dataChanged.emit(index, index, [Qt.CheckStateRole])
            self.checked_components_changed.emit()
            return True
        else:
            return False

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        default = super().flags(index)
        if index.isValid() and self.components[index.row()].enabled:
            return default
        else:
            return Qt.ItemFlags(Qt.NoItemFlags)

    @classmethod
    def component_check_state_role(cls, component: Component) -> Qt.CheckStateRole:
        if component.selected:
            return Qt.Checked
        else:
            return Qt.Unchecked

    @classmethod
    def set_component_check_state_role(cls, component: Component, check_state: Qt.CheckStateRole) -> None:
        if check_state == Qt.Checked:
            component.selected = True
        else:
            component.selected = False

    def check_index(self, index: QModelIndex) -> None:
        component = self.components[index.row()]
        if component.enabled:
            if component.selected:
                self.setData(index, Qt.Unchecked, Qt.CheckStateRole)
            else:
                self.setData(index, Qt.Checked, Qt.CheckStateRole)
                for dependency in component.dependencies:
                    list_index = self.__index_by_id.get(dependency)
                    model_index = index.sibling(list_index, 0)
                    if not self.components[list_index].selected:
                        self.check_index(model_index)

    def check_index_exclusively(self, index: QModelIndex) -> None:
        self.beginResetModel()
        selected = self.__components[index.row()]
        for component in self.__components:
            if component.enabled and component.visible:
                component.selected = component == selected
        self.endResetModel()

    def checked_components(self) -> List[Component]:
        return [component for component in self.components if component.selected]

    def showable_checked_components(self) -> List[Component]:
        return [component for component in self.checked_components() if component.visible]

    def checked_names(self) -> List[str]:
        components = self.showable_checked_components()
        names = [_(component.name) for component in components]
        list.sort(names)
        return names
