import traceback
from typing import List, Optional

from PyQt5 import uic
from PyQt5.QtCore import Qt, QEvent, QTranslator, QProcess
from PyQt5.QtGui import QKeyEvent, QIcon, QFont, QPixmap
from PyQt5.QtWidgets import QApplication, QMessageBox, QMainWindow

from astra_installer import globvars
from astra_installer.settings.autoinstall_config import AutoinstallConfigSaver
from astra_installer.settings.component import has_component_with_ident
from astra_installer.settings.config import ConfigLoader
from astra_installer.settings.manager import SettingsManager
from astra_installer.settings.systeminfo import AstraEdition
from astra_installer.utils import Utils as BackendUtils
from astra_installer_qt.components_dialog import ComponentsDialog
from astra_installer_qt.error_manager import ErrorManager
from astra_installer_qt.grub.grub_manager import GrubManager
from astra_installer_qt.date_time.date_time_manager import DateTimeManager
from astra_installer_qt.kernel.kernel_manager import KernelManager
from astra_installer_qt.help_dialog import HelpDialog
from astra_installer_qt.quit_dialog import QuitDialog
from astra_installer_qt.hostname_manager import HostnameManager
from astra_installer_qt.icon_button import IconButton
from astra_installer_qt.image_presentation_manager import ImagePresentationManager
from astra_installer_qt.installation_manager import InstallationManager
from astra_installer_qt.installer_locale_manager import InstallerLocaleManager
from astra_installer_qt.layout_switch_hotkey_manager import LayoutSwitchHotkeyManager
from astra_installer_qt.layout_switcher import LayoutSwitcher
from astra_installer_qt.license_manager import LicenseManager
from astra_installer_qt.log_manager import LogManager
from astra_installer_qt.pages_manager import PagesManager
from astra_installer_qt.mode_manager import ModeManager
from astra_installer_qt.pixmap_label import PixmapLabel
from astra_installer_qt.progress_manager import ProgressManager
from astra_installer_qt.storage.ui.wait_dialog import WaitDialog
from astra_installer_qt.system_locale_manager import SystemLocaleManager
from astra_installer_qt.timezone_manager import TimeZoneManager
from astra_installer_qt.user_manager import UserManager
from astra_installer_qt.utils import Utils
from astra_installer_qt.virtual_keyboard import VirtualKeyboard

if not globvars.oem_mode:
    from astra_installer.storage.utils.device_action import DeviceActionUtils
    from astra_installer.storage.template.storage_template import StorageTemplate
    from astra_installer_qt.storage.storage_manager import StorageManager

translate = QApplication.translate

mainwindow_ui = Utils.file_path('ui', 'mainwindow.ui')
UiFormClass, UiFormBase = uic.loadUiType(mainwindow_ui)  # type: QMainWindow, QMainWindow


class MainWindowCommon(UiFormClass, UiFormBase):
    def __init__(self, locale: str) -> None:
        super(MainWindowCommon, self).__init__()
        self.locale = locale
        self.setupUi(self)
        self.load_settings()
        self.setup_common()

    def load_settings(self):
        config = ConfigLoader().load()
        SettingsManager.create_settings(config)
        if self.locale in SettingsManager.installer_locale.locales:
            SettingsManager.installer_locale.selected = self.locale
        custom_template = StorageTemplate(name='Custom template')
        SettingsManager.storage.templates += [custom_template]

    def setup_common(self) -> None:
        if SettingsManager.systeminfo.oem:
            self.disable_close = SettingsManager.systeminfo.is_kiosk_session
        else:
            self.disable_close = False

        QApplication.setWindowIcon(QIcon(Utils.file_path('resources', 'icons', 'logo.svg')))

        astra_logo = Utils.file_path('resources', 'icons', 'astra_full_logo_white.png')
        keyboard_icon = Utils.file_path('resources', 'icons', 'input-keyboard.png')
        datetime_icon = Utils.file_path('resources', 'icons', 'setup-datetime.png')
        error_icon = Utils.file_path('resources', 'icons', 'task-attention.png')

        slides = []
        # MainWindowCommon.make_current_distr_slide_first(slides)
        images = [Utils.file_path('resources', 'slides', '{}.png'.format(name)) for name in slides]

        self.logo_pixmap = PixmapLabel.gen_pixmap(astra_logo)
        self.logo1 = PixmapLabel(self.labelLogo, self.logo_pixmap)
        self.logo2 = PixmapLabel(self.labelLogo2, self.logo_pixmap)
        self.logo3 = PixmapLabel(self.labelLogo3, self.logo_pixmap)

        # Fix for icons in the e2k-8c
        if SettingsManager.systeminfo.arch == "e2k-8c":
            license_icon = Utils.file_path('resources', 'icons', 'license_qr.png')
            language_icon = Utils.file_path('resources', 'icons', 'select-language.png')

            self.labelLicenseQR.setPixmap(QPixmap(license_icon))
            self.labelLanguageIcon.setPixmap(QPixmap(language_icon))

        self.error_icon_label = PixmapLabel(self.labelErrorIcon, PixmapLabel.gen_pixmap(error_icon))
        self.error = ErrorManager(self.labelErrorIcon, self.labelErrorText, self)

        self.pages_manager = PagesManager(self, self.error)

        self.installer_language = InstallerLocaleManager(self.comboBoxLanguage)

        self.virtual_keyboard_button = IconButton(keyboard_icon, self.pushButtonVirtualKeyboard)
        self.virtual_keyboard = VirtualKeyboard(self.pushButtonVirtualKeyboard)

        self.layout_switch = LayoutSwitchHotkeyManager(self.comboBoxKeyboardLayoutSwitch)
        self.layout_switcher = LayoutSwitcher(self.pushButtonLayoutSwitch)

        self.setup_datetime_button = IconButton(datetime_icon, self.pushButtonSetupDatetime)

        self.timezone = TimeZoneManager(self.comboBoxTimeZone)

        self.system_locale = SystemLocaleManager(self.comboBoxSystemLanguage)
        # self.comboBoxSystemLanguage.currentIndexChanged.connect(self.system_locale.apply)
        self.system_locale.locale_changed.connect(self.layout_switcher.load_layouts)

        self.security_components_dialog = ComponentsDialog(self.listViewFeatures, SettingsManager.security)

        self.os_components_dialog = ComponentsDialog(self.listViewComponents, SettingsManager.components)

        self.mode = ModeManager(self.comboBoxMode, self.labelMode,  self.security_components_dialog.components_manager)

        # Режиме ОЕМ, до тех пор, пока компоненты не загружены, кнопка "настроить" должна быть недоступна
        if SettingsManager.systeminfo.oem:
            if not SettingsManager.systeminfo.is_kiosk_session:
                self.pushButtonQuit1.setHidden(True)
                self.pushButtonQuit2.setHidden(True)

            self.pages_manager.disable_install_button()
            self.os_components_dialog.components_manager.componentsReady.connect(
                self.pages_manager.enable_install_button)
            self.security_components_dialog.components_manager.componentsReady.connect(
                self.pages_manager.enable_install_button)
            self.security_components_dialog.components_manager.componentsReady.connect(self.mode.apply)

        self.help_dialog = HelpDialog(self.help_path)
        self.quit_dialog = QuitDialog(parent=self)

        self.hostname = HostnameManager(self.lineEditHostname)
        self.hostname.error.connect(self.error.show_error)
        self.hostname.noerror.connect(self.error.hide_error)

        self.grub_manager = GrubManager(main_window=self,
                                        settings_button=self.pushButtonGrubPasswordSettings,
                                        info_label=self.labelGrubPasswordInfo,
                                        title_label=self.labelGrubPasswordSettings)
        self.user = UserManager(self.lineEditUsername, self.lineEditPassword,
                                self.lineEditPasswordConfirmation,
                                self.grub_manager)

        self.kernel = KernelManager(main_window=self,
                                    settings_button=self.pushButtonKernelSettings,
                                    title_label=self.labelKernelSettings,
                                    info_label=self.labelKernelInfo)

        self.date_time_manager = DateTimeManager(main_window=self,
                                                 settings_button=self.pushButtonSetupDatetime,
                                                 timezone=self.timezone)

        self.license = LicenseManager(self.checkBoxAgree)
        self.license.license_accepted.connect(self.on_license_accepted)
        self.comboBoxMode.currentIndexChanged.connect(self.on_mode_selected)
        self.pushButtonOpenSavedCopy.clicked.connect(self.show_license)

        # self.image_presentation_manager = ImagePresentationManager(images, self.labelImage)

        self.log_manager = LogManager(self.pushButtonOpenLog,
                                      self.installationProcessWidgets,
                                      self.plainTextEditLog,
                                      self.labelInstallationInfo,
                                      self.labelImage)

        self.progress_manager = ProgressManager(self.labelInstallationProgressDescription,
                                                self.labelPercent,
                                                self.progressBarInstallation)

        self.installation_manager = InstallationManager(self.progress_manager,
                                                        self.log_manager)

        self.installer_language.locale_changed.connect(self.layout_switch.retranslate)
        self.installer_language.locale_changed.connect(self.timezone.retranslate)
        self.installer_language.locale_changed.connect(self.system_locale.retranslate)
        self.installer_language.locale_changed.connect(self.os_components_dialog.retranslate)
        self.installer_language.locale_changed.connect(self.security_components_dialog.retranslate)
        self.installer_language.locale_changed.connect(self.help_dialog.retranslate)
        self.installer_language.locale_changed.connect(self.quit_dialog.retranslate)
        self.installer_language.locale_changed.connect(self.error.retranslate)
        self.installer_language.locale_changed.connect(self.grub_manager.retranslate)
        self.installer_language.locale_changed.connect(self.date_time_manager.retranslate)
        self.installer_language.locale_changed.connect(self.kernel.retranslate)
        self.installer_language.locale_changed.connect(self.mode.retranslate)
        self.installer_language.locale_changed.connect(self.log_manager.retranslate)

        self.user.error.connect(self.error.show_error)
        self.user.noerror.connect(self.error.hide_error)

        self.error.error.connect(self.pages_manager.disable_install_button)
        self.error.noerror.connect(self.pages_manager.enable_install_button)

        self.pushButtonInstall.clicked.connect(self.check_settings)
        self.installation_manager.finished.connect(self.installation_finished)
        self.pushButtonFinish.clicked.connect(self.finish_button_clicked)

        self.checkBoxRebootAfterInstall.stateChanged.connect(self.checkBoxRebootAfterInstall2.setCheckState)
        self.checkBoxRebootAfterInstall2.stateChanged.connect(self.on_reboot_after_install_changed)

        self.labelLicenseUrl.linkActivated.connect(self.open_uri)

        self.pushButtonQuit1.clicked.connect(self.quit)
        self.pushButtonQuit2.clicked.connect(self.quit)

        self.install_messagebox_title = translate('MainWindowCommon', 'Install confirmation')

        for button in [self.pushButtonHelp1, self.pushButtonHelp2, self.pushButtonHelp3]:
            button.clicked.connect(self.show_help)

        for widget, value in {self.labelDesc: 1.0,
                              self.labelDesc2: 1.0,
                              self.labelDesc3: 1.0,
                              self.labelRelease: 0.85,
                              self.labelRelease2: 0.85,
                              self.labelRelease3: 0.85}.items():
            self.fix_font_spacing(widget, value)

        for widget in {self.lineRelease, self.lineRelease2, self.lineRelease3}:
            widget.setHidden(True)

        self.hostname.read_settings()
        self.user.read_settings()

        self.set_license_link()
        self.pushButtonContinue.setFocus()

    def fix_font_spacing(self, widget, value: float) -> None:
        font = widget.font()
        font.setLetterSpacing(QFont.AbsoluteSpacing, value)
        widget.setFont(font)

    @property
    def help_path(self) -> str:
        return Utils.file_path('resources', 'help')

    def setup_backend(self) -> None:
        self.layout_switch.setup_backend()
        self.timezone.setup_backend()
        self.hostname.setup_backend()
        self.user.setup_backend()
        self.system_locale.apply()
        self.security_components_dialog.setup_backend()
        self.os_components_dialog.setup_backend()
        SettingsManager.disk.reboot_after_install = self.checkBoxRebootAfterInstall.isChecked()

    @staticmethod
    def make_current_distr_slide_first(slides: List[str]) -> None:
        name = SettingsManager.systeminfo.name
        index = slides.index(name)
        element = slides.pop(index)
        slides.insert(0, element)

    @property
    def install_messagebox_question(self) -> str:
        return translate('MainWindowCommon', 'Do you want to continue?')

    @property
    def os_components_text(self) -> str:
        return translate('MainWindowCommon', 'Select OS components')

    @property
    def window_title(self) -> str:
        return translate('title', 'Astra Linux Installer')

    @property
    def password_label_text(self) -> str:
        return translate('text', 'Password:')

    @property
    def reboot_checkbox_text(self) -> str:
        return translate('text', 'Reboot computer after installation is complete')

    @property
    def sucessfully_installed_text(self) -> str:
        return translate('text', 'You have successfully installed')

    @property
    def install_button_text(self) -> str:
        return translate('text', 'Install')

    def set_license_link(self) -> None:
        link = 'www.astralinux.ru/information/licenses'
        html = "<html><head/><body><a href='http://{link}'>{link}</a></body></html>".format(link=link)
        if Utils.is_wm_present():
            self.labelLicenseUrl.setText(html)
        else:
            self.labelLicenseUrl.setText(link)

    def open_uri(self, uri: str) -> None:
        # fly-open не открывает ссылки при запуске от sudo, запустим от юзера
        try:
            uid = BackendUtils.get_sudo_uid()
            QProcess.startDetached('sudo -H -u #{uid} fly-open -e {uri}'.format(uid=uid, uri=uri))
        except Exception:
            print('Failed to open URI {uri}'.format(uri=uri))
            traceback.print_exc()

    def retranslate_labels(self) -> None:
        name = translate('release', SettingsManager.systeminfo.name.upper())
        release = translate('text', 'RELEASE')
        release_name = "{release} «{name}»".format(release=release, name=name)
        for label in [self.labelRelease, self.labelRelease2, self.labelRelease3]:
            # label.setText(release_name)
            label.setHidden(True)
        # self.labelReleaseName.setText(name.capitalize())
        self.labelReleaseName.setHidden(True)
        # release_title = "<html><head/><body><p>Astra Linux<br>" \
        #                 "<span style='font-weight:600;'>{name}" \
        #                 "</span></p></body></html>".format(name=name.capitalize())
        release_title = "<html><head/><body>Astra Linux</body></html>"
        self.labelReleaseTitle2.setText(release_title)
        self.setWindowTitle(self.window_title)
        self.labelPassword.setText(self.password_label_text)
        self.checkBoxRebootAfterInstall.setText(self.reboot_checkbox_text)
        self.checkBoxRebootAfterInstall2.setText(self.reboot_checkbox_text)
        self.labelSucessfullyInstalled.setText(self.sucessfully_installed_text)
        self.pushButtonInstall.setText(self.install_button_text)

    def on_license_accepted(self, license_accepted: bool) -> None:
        if self.pages_manager.current_page() == 0:
            if self.comboBoxMode.currentIndex() >= 0:
                if license_accepted:
                    self.pages_manager.enable_continue_button()
                else:
                    self.pages_manager.disable_continue_button()
            else:
                self.pages_manager.disable_continue_button()

    def on_mode_selected(self, current_index: int) -> None:
        if self.pages_manager.current_page() == 0:
            if current_index >= 0:
                if self.checkBoxAgree.isChecked():
                    self.pages_manager.enable_continue_button()
                else:
                    self.pages_manager.disable_continue_button()
            else:
                self.pages_manager.disable_continue_button()

    def on_reboot_after_install_changed(self) -> None:
        SettingsManager.disk.reboot_after_install = self.checkBoxRebootAfterInstall2.isChecked()

    def check_settings(self) -> None:
        error_message = SettingsManager.storage.validate()
        # TODO
        # if error and error.code == 2:
        #     # Отсутствие загрузочных разделов и дисков можно игнорировать
        #     dialog = QMessageBox(parent=self)
        #     dialog.setIcon(QMessageBox.Critical)
        #     dialog.setWindowTitle(_('Disks layouts error'))
        #     dialog.setText(str(error))
        #     dialog.setStandardButtons(QMessageBox.Cancel | QMessageBox.Ignore)
        #     dialog.button(QMessageBox.Cancel).setText(_('Cancel installation'))
        #     dialog.button(QMessageBox.Ignore).setText(_('Continue anyway'))
        #     dialog.button(QMessageBox.Cancel).setFocus()
        #     if dialog.exec() == QMessageBox.Ignore:
        #         self.ask_about_install()
        #         self.activateWindow()
        if error_message:
            QMessageBox.critical(self, translate('text', 'Disks layouts error'), error_message)
        else:
            warnings = SettingsManager.storage.warnings()
            if not SettingsManager.kernel.is_enough_ram(SettingsManager.systeminfo.ram):
                warnings.append(translate('text', 'The available amount of RAM may not be enough for the correct work of the system'))
            if warnings:
                QMessageBox.warning(self, translate('text', 'Warnings'), '\n'.join(warnings))

            self.setup_backend()
            self.save_preseed_config()
            if self.check_oem_and_sudo_password():
                self.ask_about_install()
            self.activateWindow()

    def check_oem_and_sudo_password(self) -> bool:
        features = self.security_components_dialog.components_manager.model.checked_components()
        if has_component_with_ident('astra-installer-oem-control', features) \
                and has_component_with_ident('astra-sudo-control', features):
            title = translate('MainWindowCommon', 'OEM warning')
            question = translate('MainWindowCommon',
                                 'You have selected the additional settings "Enable sudo password" and "OEM installation". '
                                 'The user will not be able to run the initial system configuration utility without a password. '
                                 'Do you want to continue?')
            return QMessageBox.question(self, title, question) == QMessageBox.Yes
        else:
            return True

    def ask_about_install(self) -> None:
        title = translate('MainWindowCommon', self.install_messagebox_title)
        question = translate('MainWindowCommon', self.install_messagebox_question)
        answer = QMessageBox.question(self, title, question)
        if answer == QMessageBox.Yes:
            self.start_installation()

    def finish_button_clicked(self) -> None:
        Utils.reboot()

    def installation_finished(self) -> None:
        self.installationProcessWidgets2.setCurrentIndex(1)
        self.pushButtonFinish.setEnabled(True)
        if SettingsManager.systeminfo.oem:
            self.disable_close = SettingsManager.systeminfo.is_kiosk_session
        else:
            self.disable_close = False

    # noinspection PyPep8Naming
    def keyPressEvent(self, keyevent: QKeyEvent) -> None:
        if keyevent.key() == Qt.Key_F1:
            self.show_help()

    # noinspection PyPep8Naming
    def changeEvent(self, event: QEvent) -> None:
        if event.type() == QEvent.LanguageChange:
            self.retranslate_labels()
            self.retranslateUi(self)

    # noinspection PyPep8Naming
    def closeEvent(self, event: QEvent) -> None:
        if self.disable_close:
            event.ignore()

    def quit(self) -> None:
        self.quit_dialog.exec()
        self.activateWindow()
        # QApplication.instance().quit()

    def save_preseed_config(self) -> None:
        try:
            # Текущие настройки сохраняются как шаблон для конфига автоустановки
            config = SettingsManager.dump_selected_settings()
            AutoinstallConfigSaver.save(config)
        except Exception:
            traceback.print_exc()
            print(translate('text', 'Failed to log preseed config based on selected settings'))

    def start_installation(self) -> None:
        # self.image_presentation_manager.run()
        self.pages_manager.open_page3()
        self.disable_close = True
        self.installation_manager.install()

    def show_help(self) -> None:
        self.help_dialog.exec()
        self.activateWindow()

    def show_license(self) -> None:
        self.license.show_license()
        self.activateWindow()


class MainWindow(MainWindowCommon):
    def __init__(self, locale: str) -> None:
        super(MainWindow, self).__init__(locale)
        self.storage_manager = None  # type: Optional[StorageManager]
        self.setup()

    def setup(self) -> None:
        self.storage_manager = StorageManager(main_window=self,
                                              settings_button=self.pushButtonStorageSettings,
                                              info_label=self.labelStorageInfo)
        self.installer_language.locale_changed.connect(self.storage_manager.retranslate)
        self.installer_language.change_installer_locale()

    def ask_about_install(self) -> None:
        title = translate('MainWindowCommon', self.install_messagebox_title)
        question = translate('MainWindow', 'Do you really want to start installing the operating system?')

        dialog = QMessageBox(parent=self)
        dialog.setIcon(QMessageBox.Question)
        dialog.setWindowTitle(title)
        dialog.setText(question)
        dialog.setStandardButtons(QMessageBox.Yes | QMessageBox.No)

        all_device_actions = self.storage_manager.storage.current_template.storage_state.actions
        clean_out_disks, short_actions_list = DeviceActionUtils.short_actions_list(all_device_actions)
        text = '<br>'.join([translate('text', "All data on disk '{name}' will be destroyed").format(name=disk.name)
                            for disk in clean_out_disks])

        if short_actions_list:
            if clean_out_disks:
                text += '<br>' + translate('text', 'also irreversible actions will be made to disk:')
            else:
                text += translate('text', 'Irreversible actions will be made to disk:')
            text += '<br>'
            text += '<br>'.join([DeviceActionUtils.action_info(action) for action in short_actions_list])

        if text:
            dialog.setInformativeText('<b>{}</b>'.format(text))

        if [action for action in all_device_actions]:
            details = translate('text', 'Actions will be made to disk:') + '\n'
            details += '\n'.join([DeviceActionUtils.action_info(action) for action in all_device_actions])
            dialog.setDetailedText(details)

        if dialog.exec() == QMessageBox.Yes:
            self.start_installation()

    def check_settings(self) -> None:
        current_template = SettingsManager.storage.current_template
        if not current_template.storage_state or not current_template.storage_state_based_on_actual_data:
            wait_dialog = WaitDialog()
            report, state = Utils.wait_thread(widget=wait_dialog,
                                              method=SettingsManager.storage.init_template,
                                              args=[current_template])
            if not report.is_valid():
                QMessageBox.critical(self,
                                     translate('text', 'Disks layouts error'),
                                     '\n'.join([str(message) for message in report.get_messages()]))
                return
            else:
                SettingsManager.storage.current_template.storage_state = state

        if current_template.storage_state.partition_module.requires_passphrase():
            current_template.storage_state = None
            QMessageBox.critical(self,
                                 translate('text', 'Disks layouts error'),
                                 translate('text', 'Passphrase not entered for encrypted devices'))
            return

        super().check_settings()
