from typing import Iterable, Generator

from PyQt5.QtCore import QObject, Qt, QEvent
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QLabel


class PixmapLabel(QObject):
    @classmethod
    def gen_pixmap(cls, file: str) -> QPixmap:
        return QPixmap(file)

    @classmethod
    def gen_pixmaps(cls, files: Iterable[str]) -> Generator[QPixmap, None, None]:
        for file in files:
            yield cls.gen_pixmap(file)

    @classmethod
    def scale_pixmap(cls, label: QLabel, pixmap: QPixmap) -> QPixmap:
        width = label.width()
        height = label.height()
        return pixmap.scaled(width, height, Qt.KeepAspectRatio, Qt.SmoothTransformation)

    @classmethod
    def scale_pixmaps(cls, label: QLabel, pixmaps: Iterable[QPixmap]) -> Generator[QPixmap, None, None]:
        for pixmap in pixmaps:
            yield cls.scale_pixmap(label, pixmap)

    @classmethod
    def set_pixmap(cls, label: QLabel, pixmap: QPixmap) -> None:
        scaled_pixmap = cls.scale_pixmap(label, pixmap)
        label.setPixmap(scaled_pixmap)

    @classmethod
    def set_pixmap_from_file(cls, label: QLabel, image_file: str) -> None:
        pixmap = cls.gen_pixmap(image_file)
        cls.set_pixmap(label, pixmap)

    def __init__(self, label: QLabel, pixmap=None) -> None:
        super(PixmapLabel, self).__init__()
        self.label = label
        self.__pixmap = pixmap
        self.label.installEventFilter(self)

    @property
    def pixmap(self) -> QPixmap:
        return self.__pixmap

    @pixmap.setter
    def pixmap(self, pixmap: QPixmap) -> None:
        self.__pixmap = pixmap
        self.update_label()

    def update_label(self) -> None:
        self.set_pixmap(self.label, self.pixmap)

    # noinspection PyPep8Naming
    def eventFilter(self, qobject: QObject, event: QEvent) -> bool:
        if event.type() == QEvent.Resize:
            self.update_label()
            return True
        else:
            return False
