from typing import TYPE_CHECKING

import sys
import threading

import gi
from PyQt5.QtCore import QCoreApplication

gi.require_version("Gtk", "3.0")
gi.require_version("GLib", "2.0")
gi.require_version("Gdk", "3.0")

from gi.repository import Gtk, Gdk, GLib

import blivet.deviceaction as deviceaction
from blivet.formats import fs

from blivetgui.list_actions import ListActions
from blivetgui.actions_menu import ActionsMenu
from blivetgui.actions_toolbar import DeviceToolbar
from blivetgui.visualization.logical_view import LogicalView
from blivetgui.visualization.physical_view import PhysicalView
from blivetgui.blivetgui import BlivetGUI
from blivetgui.gui_utils import locate_css_file
from blivetgui.i18n import _ as blivet_gui_translation

from astra_installer_kickstart.installer_storage import InstallerStorage
from astra_installer_qt.storage.blivet.installer_blivet_client import InstallerBlivetGUIClient
from astra_installer_qt.storage.blivet.installer_blivet_add_dialog import InstallerBlivetAddDialog
from astra_installer_qt.storage.blivet.installer_list_partitions import InstallerListPartition
from astra_installer_qt.storage.blivet.installer_list_parents import InstallerListParents
from astra_installer_qt.storage.blivet.installer_list_devices import InstallerListDevices
from astra_installer_qt.storage.blivet.installer_disk_actions_menu import InstallerDiskActionsMenu
from astra_installer_qt.storage.blivet.installer_disk_info_dialog import InstallerDiskInformationDialog
from astra_installer_qt.storage.blivet.installer_actions_toolbar import InstallerActionsToolbar
from astra_installer_qt.storage.blivet.installer_main_menu import InstallerMainMenu
from astra_installer_qt.storage.ui.storage_settings_dialog import StorageSettingsDialog
from astra_installer_qt.storage.blivet.installer_blivet_gui_exception_handler import InstallerBlivetGUIExceptionHandler
from astra_installer_qt.utils import Utils as QtUtils

translate = QCoreApplication.translate

if TYPE_CHECKING:
    from gettext import gettext as _

# Hack for correct retranslation
action_type_descriptions = {
    deviceaction.ACTION_TYPE_NONE: {
        deviceaction.ACTION_OBJECT_NONE: ""
    },
    deviceaction.ACTION_TYPE_DESTROY: {
        deviceaction.ACTION_OBJECT_DEVICE: "destroy device",
        deviceaction.ACTION_OBJECT_FORMAT: "destroy format"
    },
    deviceaction.ACTION_TYPE_RESIZE: {
        deviceaction.ACTION_OBJECT_DEVICE: "resize device",
        deviceaction.ACTION_OBJECT_FORMAT: "resize format"
    },
    deviceaction.ACTION_TYPE_CREATE: {
        deviceaction.ACTION_OBJECT_DEVICE: "create device",
        deviceaction.ACTION_OBJECT_FORMAT: "create format"
    },
    deviceaction.ACTION_TYPE_ADD: {
        deviceaction.ACTION_OBJECT_CONTAINER: "add container member"
    },
    deviceaction.ACTION_TYPE_REMOVE: {
        deviceaction.ACTION_OBJECT_CONTAINER: "remove container member"
    },
    deviceaction.ACTION_TYPE_CONFIGURE: {
        deviceaction.ACTION_OBJECT_DEVICE: "configure device", # There is bug in the blivet
        deviceaction.ACTION_OBJECT_FORMAT: "configure format"
    }
}

class InstallerBlivetGUI(BlivetGUI):
    installer_mode = True
    allow_ignore = True

    __action_types = {'add': [deviceaction.ACTION_TYPE_CREATE, deviceaction.ACTION_TYPE_ADD],
                      'delete': [deviceaction.ACTION_TYPE_DESTROY, deviceaction.ACTION_TYPE_REMOVE],
                      'edit': [deviceaction.ACTION_TYPE_RESIZE, deviceaction.ACTION_TYPE_CONFIGURE]}
    __misc_type = 'misc'

    def __init__(self, settings_dialog: StorageSettingsDialog):
        self.client = InstallerBlivetGUIClient()
        self.__settings_dialog = settings_dialog

        self.builder = Gtk.Builder()
        self.builder.set_translation_domain("blivet-gui")
        self.builder.add_from_file(QtUtils.file_path('storage', 'ui', 'blivet-gui.ui'))

        self.ignored_disks = []

        # allow creating ntfs filesystem
        # XXX: This shouldn't be necessary, NTFS is already "_formattable",
        # I don't have idea what "_supported" is for, the "supported" property
        # depends on this and "utils_available" so it is False when "ntfsprogs"
        # are not installed and that should be enough.
        fs.NTFS._supported = True

        # supported filesystems
        self._supported_filesystems = []

        # CSS styles
        css_provider = Gtk.CssProvider()
        css_provider.load_from_path(locate_css_file("rectangle.css"))
        screen = Gdk.Screen.get_default()  # pylint: disable=no-value-for-parameter
        style_context = Gtk.StyleContext()
        style_context.add_provider_for_screen(screen, css_provider, Gtk.STYLE_PROVIDER_PRIORITY_USER)

        # MainWindow
        self.main_window = self.builder.get_object("main_window")
        self.main_window.set_resizable(False)
        self.main_window.connect("delete-event", self.quit)

        # Exception handling
        self.exc = InstallerBlivetGUIExceptionHandler(self.main_window, sys.excepthook)
        self.exc.allow_ignore = False  # don't allow to ignore exceptions right now
        sys.excepthook = self.exc.handle_exception

        # BlivetUtils
        # self.blivet_init()

        # Atexit
        # atexit.register(self.client.quit)

        #Paned
        self.paned = self.builder.get_object("paned")

        # MainMenu
        self.main_menu = InstallerMainMenu(self)

        # ActionsMenu
        self.popup_menu = ActionsMenu(self)
        self.disk_popup_menu = InstallerDiskActionsMenu(self)

        # DeviceToolbar
        self.device_toolbar = DeviceToolbar(self)

        # ActionsToolbar
        self.actions_toolbar = InstallerActionsToolbar(self)
        vbox_right = self.builder.get_object("vbox_right")
        vbox_right.pack_start(self.actions_toolbar.toolbar, expand=False, fill=True, padding=0)

        # ListDevices
        self.list_devices = InstallerListDevices(self)

        # ListPartitions
        self.list_partitions = InstallerListPartition(self)

        # ListParents
        self.list_parents = InstallerListParents(self)

        # ListActions
        self.label_actions = self.builder.get_object("label_actions")
        self.button_actions = self.builder.get_object("button_actions")
        self.button_actions.connect("clicked", self.show_actions)
        self.list_actions = ListActions(self)

        # Vizualisation
        self.logical_view = LogicalView(self)
        self.builder.get_object("image_window").add(self.logical_view.hbox)

        self.physical_view = PhysicalView(self)
        self.builder.get_object("scrolledwindow_physical").add(self.physical_view.vbox)

        # allow ignoring exceptions
        self.exc.allow_ignore = True

        # set some defaults from blivet now
        # config.default_fstype = self.client.remote_call("get_default_filesystem")

        self.initialize()

    def set_background_color(self, color_name: str) -> None:
        state = Gtk.StateType.NORMAL
        color = Gdk.color_parse(color_name)
        self.main_window.modify_bg(state, color)
        self.builder.get_object('notebook_views').modify_bg(state, color)

    @classmethod
    def action_type(cls, type) -> str:
        try:
            for str_type, types in cls.__action_types.items():
                if type in types:
                    return str_type
        except KeyError:
            pass

        return cls.__misc_type

    def load_storage(self, storage: InstallerStorage) -> None:
        self.client.set_storage(storage)
        self.list_devices.load_devices()
        for action in storage.devicetree.actions:
            action.type_desc_str = _(action_type_descriptions[action.type][action.obj]) # Hack for correct retranslation
            self.list_actions.append(action_type=self.action_type(action.type),
                                     action_desc=action.type_desc_str,
                                     blivet_actions=[action])
        self.list_devices.disks_view.set_cursor(1)
        self.list_devices.disks_window.set_visible(len(self.list_devices.device_list) > 2)
        # hack for correct width
        if self.paned.get_screen().get_width() <= 800:
            self.paned.set_position(140)

    def unload_storage(self) -> None:
        self.client.set_storage(None)
        self.list_actions.clear()
        self.list_partitions.clear()
        self.list_devices.clear()

        # don't allow to ignore exceptions raised during reset
        self.exc.allow_ignore = False

        self._handle_user_change()
        self.list_devices.update_devices_view()
        self.update_partitions_view()

        # allow ignoring exceptions now
        self.exc.allow_ignore = True

    @property
    def supported_filesystems(self):
        self._supported_filesystems = self.client.remote_call("get_supported_filesystems",
                                                              self.installer_mode,
                                                              self.__settings_dialog.use_bios)
        return self._supported_filesystems

    def initialize(self):
        # self.list_devices.load_devices()
        self.list_actions.initialize()

        # select first device in ListDevice
        self.list_devices.disks_view.set_cursor(1)
        self.main_window.show()
        self.list_devices.disks_view.set_cursor(0)
        self.list_devices.disks_window.set_visible(False)

    def update_physical_view(self):
        self.list_parents.update_parents_list(self.list_devices.selected_device)
        self.physical_view.visualize_parents(self.list_parents.parents_list)

        if self.list_devices.selected_device and self.list_devices.selected_device.is_disk:
            self._set_physical_view_visible(False)
        else:
            self._set_physical_view_visible(True)

    def add_device(self, _widget=None):
        """ Show dialog for adding new device and create the device based on
            user selection """

        selected_device = self.list_partitions.selected_partition[0]

        # allow adding new device?
        allow, msg = self._allow_add_device(selected_device)

        if not allow:
            self.show_error_dialog(msg)
            return

        # uninitialized disk or mdarray -> add a disklabel
        if selected_device.type == "free space":
            if selected_device.is_uninitialized_disk:
                self._add_disklabel(disk=selected_device.disk)
                return
            elif selected_device.parents[0].type == "mdarray" and not selected_device.parents[0].format.type:
                self._add_disklabel(disk=selected_device.parents[0])
                return

        # adding a new device is allowed when selected both free space and some
        # "normal" devices -- we need both information: "future" parent and
        # selected free space (if available)
        if selected_device.type == "free space":
            selected_parent = selected_device.parents[0]
            selected_free = selected_device
        else:
            selected_parent = selected_device
            selected_free = self.client.remote_call("get_free_device", selected_device)

        if self.installer_mode:
            mountpoints = self.client.remote_call("get_mountpoints")
        else:
            mountpoints = []

        dialog = InstallerBlivetAddDialog(parent_window=self.main_window,
                                      selected_parent=selected_parent,
                                      selected_free=selected_free,
                                      available_free=self.client.remote_call("get_free_info"),
                                      supported_filesystems=self.supported_filesystems,
                                      mountpoints=mountpoints,
                                      installer_mode=self.installer_mode)
        dialog.set_decorated(True)

        response = self.run_dialog(dialog)
        message = blivet_gui_translation("Failed to add the device:")

        if response == Gtk.ResponseType.OK:
            user_input = dialog.get_selection()
            result = self.client.remote_call("add_device", user_input)

            if not result.success:
                if not result.exception:
                    self.show_error_dialog(result.message)
                else:
                    self._reraise_exception(result.exception, result.traceback, message,
                                            dialog_window=dialog)

            else:
                if result.actions:
                    action_str = blivet_gui_translation("add {size} {type} device") \
                        .format(size=str(user_input.size_selection.total_size),
                                type=user_input.device_type)

                    self.list_actions.append("add", action_str, result.actions)

            self._handle_user_change()
            self.list_devices.update_devices_view()
            self.update_partitions_view()

        dialog.destroy()

    def reload(self, _widget=None):
        """ Reload storage information

            :param widget: widget calling this function (only for calls via signal.connect)
            :type widget: Gtk.Widget()

        """

        title = translate('text', 'Confirm reload storage')
        msg = translate('text', 'Are you sure you want reload storage?')
        response = self.show_confirmation_dialog(title, msg)
        if not response:
            return

        self.__settings_dialog.init_template(force_update=True)

    def perform_actions(self, dialog):
        """ Perform queued actions
        """

        def end(success, error, traceback):
            if success:
                try:
                    dialog.stop()
                except:
                    pass
            else:
                message = blivet_gui_translation("Failed to perform the actions:")
                dialog.destroy()
                self.main_window.set_sensitive(False)
                self._reraise_exception(error, traceback, message, dialog_window=dialog)  # pylint: disable=raising-bad-type

        def show_progress(message):
            dialog.progress_msg(message)

        def do_it():
            """ Run blivet.doIt()
            """

            result = self.client.remote_do_it(lambda message: GLib.idle_add(show_progress, message))
            if result.success:
                GLib.idle_add(end, True, None, None)
            else:
                self.client.remote_call("blivet_reset")
                GLib.idle_add(end, False, result.exception, result.traceback)

            return

        # don't allow to ignore exceptions raised during do_it
        self.exc.allow_ignore = False

        thread = threading.Thread(target=do_it)
        thread.start()
        dialog.start()
        thread.join()

        self.list_actions.clear()

        self.list_devices.update_devices_view()
        self.update_partitions_view()

        # allow ignoring exceptions now
        self.exc.allow_ignore = True

    def run_dialog(self, dialog):
        try:
            self.__settings_dialog.setEnabled(False)
            return super().run_dialog(dialog)
        finally:
            self.__settings_dialog.setEnabled(True)

    def disk_information(self, _widget=None):
        """ Display information about currently selected disk
        """

        blivet_device = self.list_devices.selected_device

        dialog = InstallerDiskInformationDialog(self.main_window, blivet_device)
        self.run_dialog(dialog)
        dialog.destroy()

    def set_formatted_manually(self):
        self.__settings_dialog.current_template.storage_state.is_formatted_manually = True
