#
# Copyright (C) 2014  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Vojtech Trefny <vtrefny@redhat.com>
#
#

from blivetgui.list_partitions import ListPartitions


class InstallerListPartition(ListPartitions):
    def update_partitions_list(self, selected_device):
        """ Update partition view with selected disc children (partitions)

            :param selected_device: selected device from list (eg. disk or VG)
            :type device_name: blivet.Device

        """

        self.partitions_list.clear()

        def _get_real_child(child):
            """ When adding a child device, we actually might want to add one
                of its children instead -- e.g. when adding a partition that
                is a PV we want to add the VG ("group device") etc.
            """
            if self._is_group_device(child):
                return self.blivet_gui.client.remote_call("get_group_device", child)
            elif child.format and child.format.type in ("luks", "integrity") and child.children:
                return self.blivet_gui.client.remote_call("get_luks_device", child)
            else:
                return child

        def _add_chilren(childs, parent_iter=None):
            for child in childs:
                if child.children:
                    child_iter = self._add_to_store(child, parent_iter)
                    _add_chilren(self.blivet_gui.client.remote_call("get_children", child), child_iter)

                else:
                    new_child = _get_real_child(child)
                    self._add_to_store(new_child, parent_iter)

        if selected_device:
            if selected_device.is_disk:
                childs = self.blivet_gui.client.remote_call("get_disk_children", selected_device)
                for child in childs.partitions:
                    child = _get_real_child(child)
                    child_iter = self._add_to_store(child)
                    if hasattr(child, "is_extended") and child.is_extended:
                        for logical in childs.logicals:
                            logical = _get_real_child(logical)
                            self._add_to_store(logical, child_iter)

            # lvmvg always has some children, at least a free space
            elif selected_device.type == "lvmvg":
                childs = self.blivet_gui.client.remote_call("get_children", selected_device)
                _add_chilren(childs, None)

            # for btrfs volumes and mdarrays its necessary to add the device itself to the view
            # because these devices don't need to have children (only btrfs volume or only mdarray
            # is a valid, usable device)
            elif selected_device.type == "btrfs volume" or (
                    selected_device.type == "mdarray" and not selected_device.children):
                parent_iter = self._add_to_store(selected_device)
                childs = self.blivet_gui.client.remote_call("get_children", selected_device)
                _add_chilren(childs, parent_iter)

            else:
                childs = self.blivet_gui.client.remote_call("get_children", selected_device)
                _add_chilren(childs, None)

        # select first line in partitions view
        self.select.select_path("0")
        # expand all expanders
        self.partitions_view.expand_all()

    def clear(self) -> None:
        self.partitions_list.clear()
        self.selected_partition = None
