import traceback
from typing import Optional, List, TYPE_CHECKING

from PyQt5 import uic
from PyQt5.QtCore import Qt, QCoreApplication, QProcess, QSize, QModelIndex, QObject
from PyQt5.QtGui import QGuiApplication, QCloseEvent, QStandardItemModel, QStandardItem, QPainter
from PyQt5.QtWidgets import QPushButton, QStyle, QCheckBox, QSpacerItem, QSizePolicy, QLabel, QStyleOptionViewItem
from PyQt5.QtWidgets import QWidget, QDialog, QGroupBox, QComboBox, QPushButton, QMessageBox, QItemDelegate
from pyudev import Device

from astra_installer.settings.manager import SettingsManager
from astra_installer.storage.storage import Storage
from astra_installer.storage.template.storage_template import StorageTemplate, TemplateType
from astra_installer.storage.utils.pyudev_utils import device_size
from astra_installer.storage.utils.storage import sort_devices
from astra_installer.utils import Utils as BackendUtils
from astra_installer_qt.storage.blivet.blivet_gui_widget import BlivetGuiWidget
from astra_installer_qt.storage.ui.passphrase_dialog import PassphraseDialog
from astra_installer_qt.storage.ui.wait_dialog import WaitDialog
from astra_installer_qt.utils import Utils
from astra_installer_qt.utils import Utils as QtUtils

if TYPE_CHECKING:
    from gettext import gettext as _

translate = QCoreApplication.translate

dialog_ui = QtUtils.file_path('storage', 'ui', 'storage_settings_dialog.ui')
UiFormClass, UiFormBase = uic.loadUiType(dialog_ui)  # type: QDialog, QDialog


class ComboBoxDeligate(QItemDelegate):
    def __init__(self, size_hint: QSize, parent: QObject = None):
        super().__init__(parent=parent)
        self.__size_hint = size_hint

    def paint(self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex) -> None:
        type = index.data(Qt.AccessibleDescriptionRole)
        if not type:
            childOption = option
            try:
                indent = option.fontMetrics.horizontalAdvance(" " * 4)
            except:
                indent = option.fontMetrics.boundingRect(" ").width() * 4
            childOption.rect.adjust(indent, 0, 0, 0)
            super().paint(painter, childOption, index)
        else:
            super().paint(painter, option, index)

    def sizeHint(self, option: QStyleOptionViewItem, index: QModelIndex) -> QSize:
        return self.__size_hint


class StorageSettingsDialog(UiFormClass, UiFormBase):
    def __init__(self, storage: Storage, parent: QWidget = None) -> None:
        super().__init__(parent=parent)
        self._disk_combo_box = None  # type: Optional[QComboBox]
        self._template_combo_box = None  # type: Optional[QComboBox]
        self._template_comment_label = None  # type: Optional[QLabel]
        self._details_push_button = None  # type: Optional[QPushButton]
        self._details_widget = None  # type: Optional[QWidget]
        self._use_bios_check_box = None  # type: Optional[QCheckBox]
        self._use_gpt_check_box = None  # type: Optional[QCheckBox]
        self._blivet_widget = None  # type: Optional[QWidget]
        self._close_push_button = None  # type: Optional[QPushButton]
        self._group_box = None  # type: Optional[QGroupBox]
        self.__tamplates_model = QStandardItemModel()
        self.__blivet_gui_widget = BlivetGuiWidget(settings_dialog=self)
        self.__vertical_spacer = QSpacerItem(0, 0, QSizePolicy.Fixed, QSizePolicy.Expanding)
        self.__storage = storage
        self.setupUi(self)
        self.__setup()

    @staticmethod
    def device_string(device: Device) -> str:
        count_bytes = device_size(device)
        gb_size = round(count_bytes / 1024 / 1024 / 1024, 2)

        text = '{size} {gb_text}'.format(size=gb_size, gb_text=_('Gb'))
        model = device.properties.get('ID_MODEL')
        if model:
            text += ' {model}'.format(model=model)
        text += ' ' + device.device_node

        return text

    @property
    def current_template(self) -> Optional[StorageTemplate]:
        return self._template_combo_box.currentData()

    @property
    def current_device(self) -> Optional[Device]:
        return self._disk_combo_box.currentData() if self._disk_combo_box.isEnabled() else None

    @property
    def use_bios(self) -> bool:
        return self._use_bios_check_box.isChecked() and SettingsManager.systeminfo.bios_compat_mode_supported

    @property
    def use_gpt(self) -> bool:
        return bool(self.current_template.config_file
                    and self.current_template.config_file.gpt_support
                    and self._use_gpt_check_box.isChecked())

    def retranslate(self, locale: str) -> None:
        self.retranslateUi(self)
        self.retranslate_template_combo_box()
        self.retranslate_disk_combo_box()
        self._template_comment_label.setText(_(self.current_template.description))
        self.__blivet_gui_widget.retranslate(locale)
        self.adjustSize()
        self.setGeometry(QStyle.alignedRect(Qt.LeftToRight,
                                            Qt.AlignCenter,
                                            self.size(),
                                            QGuiApplication.instance().desktop().availableGeometry()))

    def retranslate_template_combo_box(self) -> None:
        for template in self.__storage.templates:
            index = self._template_combo_box.findData(template, Qt.UserRole)
            if index < 0:
                continue
            self._template_combo_box.setItemText(index, _(template.name))
            if template.description:
                self._template_combo_box.setItemData(index, _(template.description), Qt.ToolTipRole)

        index = self._template_combo_box.findData("recomended", Qt.AccessibleDescriptionRole)
        if index >= 0:
            self._template_combo_box.setItemText(index, _("Recomended:"))
        index = self._template_combo_box.findData("other", Qt.AccessibleDescriptionRole)
        if index >= 0:
            self._template_combo_box.setItemText(index, _("Other:"))
        self._template_combo_box.setToolTip(self._template_combo_box.currentText())

    def retranslate_disk_combo_box(self) -> None:
        for device in sort_devices(self.__storage.all_physical_devices()):
            text = self.device_string(device)
            index = self._disk_combo_box.findData(device)
            if index < 0:
                continue
            self._disk_combo_box.setItemText(index, text)
        self._disk_combo_box.setToolTip(self._disk_combo_box.currentText())

    def accept(self) -> None:
        if not self.isEnabled():
            return

        if self.init_template(force_out=True):
            super().accept()

    def reject(self) -> None:
        self.accept()

    def closeEvent(self, event: QCloseEvent) -> None:
        event.ignore()
        self.reject()

    def show(self) -> None:
        active_window = QCoreApplication.instance().activeWindow()
        self.setVisible(True)
        self.exec()
        if active_window:
            active_window.activateWindow()

    def exec(self) -> int:
        self.init_template()
        return super().exec()

    def load(self) -> None:
        self.__update_devices()

        self._template_combo_box.clear()

        recomended = []
        other = []
        for template in self.__storage.templates:
            item = QStandardItem(_(template.name))
            item.setData(template, Qt.UserRole)
            if template.description:
                item.setData(_(template.description), Qt.ToolTipRole)
            if template.recomended:
                recomended.append(item)
            else:
                other.append(item)

        if recomended:
            item = QStandardItem(_("Recomended:"))
            item.setData("recomended", Qt.AccessibleDescriptionRole)
            item.setFlags((Qt.NoItemFlags))
            font = item.font()
            font.setBold(True)
            item.setFont(font)
            self.__tamplates_model.appendRow(item)
            for itm in recomended:
                self.__tamplates_model.appendRow(itm)
        if other:
            item = QStandardItem(_("Other:"))
            item.setData("other", Qt.AccessibleDescriptionRole)
            item.setFlags((Qt.NoItemFlags))
            font = item.font()
            font.setBold(True)
            item.setFont(font)
            self.__tamplates_model.appendRow(item)
            for itm in other:
                self.__tamplates_model.appendRow(itm)

        if self._template_combo_box.count() > 1:
            for index in range(self._template_combo_box.count()):
                template = self._template_combo_box.itemData(index, Qt.UserRole)
                if template:
                    if template.is_default:
                        self._template_combo_box.setCurrentIndex(index)
                        return
            self._template_combo_box.setCurrentIndex(1)

    def __setup(self) -> None:
        self.setModal(True)
        self._blivet_widget.layout().addWidget(self.__blivet_gui_widget)
        self.__hide_details_widgets()

        self._template_combo_box.setModel(self.__tamplates_model)
        self._template_combo_box.setItemDelegate(ComboBoxDeligate(self._template_combo_box.sizeHint()))

        self._close_push_button.clicked.connect(self.reject)
        self._template_combo_box.currentIndexChanged[int].connect(self.__layout_combo_box_current_index_changed)
        self._disk_combo_box.currentIndexChanged.connect(self.__disk_combo_box_current_index_changed)
        self._details_push_button.clicked.connect(self.__details_push_button_clicked)
        self._use_gpt_check_box.stateChanged.connect(self.__use_gpt_check_box_state_changed)
        self._use_bios_check_box.stateChanged.connect(self.__use_bios_check_box_state_changed)

        self.__storage.current_template_changed.connect(self.__storage_current_template_changed)
        self.__storage.storage_templates_changed.connect(self.__storage_storage_templates_changed)

        self._template_comment_label.linkActivated.connect(self.open_uri)

        self._use_bios_check_box.setVisible(SettingsManager.systeminfo.bios_compat_mode_supported)
        self._use_bios_check_box.setEnabled(SettingsManager.systeminfo.bios_compat_mode_supported)

    def __storage_current_template_changed(self, template: Optional[StorageTemplate]) -> None:
        self._template_combo_box.setCurrentIndex(self._template_combo_box.findData(template))

    def __storage_storage_templates_changed(self, _: List[StorageTemplate]) -> None:
        self.load()

    def __details_push_button_clicked(self, _: bool) -> None:
        self.__hide_details_widgets(not self.__details_widgets_is_hidden())

    def __use_gpt_check_box_state_changed(self, __: Qt.CheckState) -> None:
        self.current_template.gpt_is_used = self.use_gpt
        self.init_template()

    def __use_bios_check_box_state_changed(self, __: Qt.CheckState) -> None:
        self.current_template.install_into_bios = self.use_bios
        self.init_template()

    def __details_widgets_is_hidden(self) -> bool:
        return self._details_widget.isHidden()

    def __hide_details_widgets(self, hide: bool = True) -> None:
        self._details_widget.setHidden(hide)
        if self._details_widget.isHidden():
            details_button_text = translate('text', 'Show details')
            self.layout().insertItem(2, self.__vertical_spacer)
            self.setMinimumSize(self.minimumSize().width(), 0)
            self.resize(self.width(), 0)  # Hack for mwm
        else:
            details_button_text = translate('text', 'Hide details')
            self.layout().removeItem(self.__vertical_spacer)
            self.setMinimumSize(self.minimumSize().width(), 600)
        self._details_push_button.setText(details_button_text)
        self.adjustSize()
        self.setGeometry(QStyle.alignedRect(Qt.LeftToRight,
                                            Qt.AlignCenter,
                                            self.size(),
                                            QGuiApplication.instance().desktop().availableGeometry()))

    def __update_devices(self) -> None:
        old_device = self.current_device

        self._disk_combo_box.blockSignals(True)
        self._disk_combo_box.clear()
        for number, device in enumerate(sort_devices(self.__storage.all_physical_devices())):
            text = self.device_string(device)
            self._disk_combo_box.addItem(text, device)

        index_changed = False
        for index in range(self._disk_combo_box.count()):
            if old_device == self._disk_combo_box.itemData(index):
                index_changed = self._disk_combo_box.currentIndex() != index
                self._disk_combo_box.setCurrentIndex(index)
                break
        self._disk_combo_box.blockSignals(False)

        if not index_changed:
            self.__disk_combo_box_current_index_changed(self._disk_combo_box.currentIndex())

    def __layout_combo_box_current_index_changed(self, __: int) -> None:
        if not self.current_template:
            return

        self.__storage.current_template = self.current_template
        self._disk_combo_box.setEnabled(self.current_template.type == TemplateType.Auto)
        self.current_template.usable_device = self.current_device
        self.current_template.gpt_is_used = self.use_gpt
        self.current_template.install_into_bios = self.use_bios
        self._template_combo_box.setToolTip(self._template_combo_box.currentText())

        if self.current_template.description:
            self._template_comment_label.setVisible(True)
            self._template_comment_label.setText(_(self.current_template.description))
        else:
            self._template_comment_label.setVisible(False)

        self._use_gpt_check_box.setEnabled(bool(self.current_template.config_file
                                                and self.current_template.config_file.gpt_support))
        self._use_gpt_check_box.setVisible(bool(self.current_template.config_file
                                                and self.current_template.config_file.gpt_support))
        if self._use_gpt_check_box.isEnabled():
            use_gpt_tool_tip = ''
        else:
            use_gpt_tool_tip = translate('text', 'Current template not support gpt partition table')
        self._use_gpt_check_box.setToolTip(use_gpt_tool_tip)
        self.init_template()

        self.adjustSize()
        self.setGeometry(QStyle.alignedRect(Qt.LeftToRight,
                                            Qt.AlignCenter,
                                            self.size(),
                                            QGuiApplication.instance().desktop().availableGeometry()))

        if not self.current_template.config_file and self.__details_widgets_is_hidden():
            self.__hide_details_widgets(False)

    def __disk_combo_box_current_index_changed(self, __: int) -> None:
        self._disk_combo_box.setToolTip(self._disk_combo_box.currentText())

        if not self.current_template:
            return

        self.current_template.usable_device = self.current_device
        self.init_template()

    def init_template(self, force_out: bool = False, force_update: bool = False) -> bool:  # TODO
        if not self.isVisible() and not force_update:
            return True

        if self.__blivet_gui_widget.blivet_gui.client.get_storage():
            self.__blivet_gui_widget.blivet_gui.unload_storage()

        if self.current_template.storage_state_based_on_actual_data and not force_update:
            self.__blivet_gui_widget.blivet_gui.load_storage(
                self.current_template.storage_state.partition_module.storage)

            if force_out:
                error = self.current_template.storage_state.validate()
                if error:
                    if force_out:
                        dialog = QMessageBox(parent=self)
                        dialog.setIcon(QMessageBox.Critical)
                        dialog.setWindowTitle(translate('text', 'Configuration error'))
                        dialog.setText(error)
                        dialog.setStandardButtons(QMessageBox.Close | QMessageBox.Ignore)
                        dialog.button(QMessageBox.Close).setText(translate('text', 'Return'))
                        dialog.button(QMessageBox.Ignore).setText(translate('text', 'Force quit'))
                        dialog.button(QMessageBox.Close).setFocus()

                        exit_code = dialog.exec()
                        if exit_code == QMessageBox.Ignore:
                            return True
                    else:
                        QMessageBox.critical(self, translate('text', 'Configuration error'), error)
                    return False

                warnings = self.current_template.storage_state.warnings()
                if warnings:
                    QMessageBox.warning(self, translate('text', 'Disks layouts warnings'), '\n'.join(warnings))
            return True

        self.current_template.storage_state = None

        if not self._disk_combo_box.count():
            if not force_out:
                QMessageBox.critical(self,
                                     translate('text', 'Configuration error'),
                                     translate('text', 'No storage devices found'))
            return True

        wait_dialog = WaitDialog()
        report, state = Utils.wait_thread(widget=wait_dialog,
                                          method=self.__storage.init_template,
                                          args=[self.current_template])

        if not report.is_valid():
            error_text = '\n'.join([str(message) for message in report.get_messages()])
            if force_out:
                dialog = QMessageBox(parent=self)
                dialog.setIcon(QMessageBox.Critical)
                dialog.setWindowTitle(translate('text', 'Configuration error'))
                dialog.setText(error_text)
                dialog.setStandardButtons(QMessageBox.Close | QMessageBox.Ignore)
                dialog.button(QMessageBox.Close).setText(translate('text', 'Return'))
                dialog.button(QMessageBox.Ignore).setText(translate('text', 'Force quit'))
                dialog.button(QMessageBox.Close).setFocus()

                exit_code = dialog.exec()
                if exit_code == QMessageBox.Ignore:
                    return True
            else:
                QMessageBox.critical(self, translate('text', 'Configuration error'), error_text)
            return False

        warnings = state.warnings()
        if warnings:
            QMessageBox.warning(self, translate('text', 'Disks layouts warnings'), '\n'.join(warnings))

        if state.partition_module.requires_passphrase():
            if force_out:
                dialog = QMessageBox(parent=self)
                dialog.setIcon(QMessageBox.Critical)
                dialog.setWindowTitle(translate('text', 'Passphrase not entered'))
                dialog.setText(translate('text', 'Do you want to enter passphrase?'))
                dialog.setStandardButtons(QMessageBox.Close | QMessageBox.Ignore)
                dialog.button(QMessageBox.Close).setText(translate('text', 'Enter'))
                dialog.button(QMessageBox.Ignore).setText(translate('text', 'Force quit'))
                dialog.button(QMessageBox.Close).setFocus()

                if dialog.exec() == QMessageBox.Ignore:
                    return True

            self.current_template.storage_state = None
            dialog = PassphraseDialog(parent=self)
            if dialog.exec():
                wait_dialog = WaitDialog()
                report, state = Utils.wait_thread(widget=wait_dialog,
                                                  method=self.__storage.init_template,
                                                  args=[self.current_template, dialog.passphrases])
            else:
                return False

            if not report.is_valid():
                error_text = '\n'.join([str(message) for message in report.get_messages()])
                if force_out:
                    dialog = QMessageBox(parent=self)
                    dialog.setIcon(QMessageBox.Critical)
                    dialog.setWindowTitle(translate('text', 'Configuration error'))
                    dialog.setText(error_text)
                    dialog.setStandardButtons(QMessageBox.Close | QMessageBox.Ignore)
                    dialog.button(QMessageBox.Close).setText(translate('text', 'Return'))
                    dialog.button(QMessageBox.Ignore).setText(translate('text', 'Force quit'))
                    dialog.button(QMessageBox.Close).setFocus()

                    exit_code = dialog.exec()
                    if exit_code == QMessageBox.Ignore:
                        return True
                else:
                    QMessageBox.critical(self, translate('text', 'Configuration error'), error_text)
                return False

            warnings = state.warnings()
            if warnings:
                QMessageBox.warning(self, translate('text', 'Disks layouts warnings'), '\n'.join(warnings))

        self.__blivet_gui_widget.blivet_gui.load_storage(self.current_template.storage_state.partition_module.storage)

        return True

    def open_uri(self, uri: str) -> None:
        # fly-open не открывает ссылки при запуске от sudo, запустим от юзера
        try:
            uid = BackendUtils.get_sudo_uid()
            QProcess.startDetached('sudo -H -u #{uid} fly-open -e {uri}'.format(uid=uid, uri=uri))
        except Exception:
            print('Failed to open URI {uri}'.format(uri=uri))
            traceback.print_exc()

    def adjustSize(self):
        super().adjustSize()
        # Hack for correct height when the fixed vertical size policy ignored
        if self._template_comment_label.height() > 22 \
                and (not self.isVisible() or self._template_comment_label.isVisible()):
            self._group_box.setMinimumHeight(105)
            if self.minimumHeight() < 600:
                self.setMinimumHeight(165)
                self.resize(self.width(), 165)  # Hack for mwm
        else:
            self._group_box.setMinimumHeight(0)
        # Hack for correct width
        if QGuiApplication.primaryScreen().size().width() > 800:
            self.setMinimumWidth(916)
        else:
            self._template_combo_box.setMaximumWidth(247)
        super().adjustSize()
