from typing import TYPE_CHECKING

from PyQt5.QtCore import QObject, pyqtSignal
from PyQt5.QtWidgets import QLineEdit

from astra_installer.settings.manager import SettingsManager
from astra_installer_qt.grub.grub_manager import GrubManager

if TYPE_CHECKING:
    from gettext import gettext as _


class UserManager(QObject):
    error = pyqtSignal(str, str)  # id, text
    noerror = pyqtSignal(str)  # id

    def __init__(self, lineedit_username: QLineEdit, lineedit_password: QLineEdit,
                 lineedit_password_confirmation: QLineEdit, grub_manager: GrubManager) -> None:
        super(UserManager, self).__init__()
        self.lineedit_username = lineedit_username
        self.lineedit_password = lineedit_password
        self.lineedit_password_confirmation = lineedit_password_confirmation
        self.grub_manager = grub_manager
        self.backend = SettingsManager.user
        self.invalid_username = 'username'
        self.invalid_password = 'password'
        self.error_style = "background: #FFFFFF;" \
                           "border: 1px solid #FF0000;" \
                           "border-radius: 2px;" \
                           "border-radius: 2px;"
        self.setup()

    def setup(self) -> None:
        self.lineedit_username.textChanged.connect(self.on_lineedit_username_text_changed)
        self.lineedit_password.textChanged.connect(self.on_lineedit_password_text_changed)
        self.lineedit_password_confirmation.textChanged.connect(self.on_lineedit_password_text_changed)

    def setup_backend(self) -> None:
        self.backend.username = self.lineedit_username.text()
        self.backend.password = self.lineedit_password.text()
        self.backend.setup_grub_password = self.grub_manager.setup_grub_password
        if self.grub_manager.grub_password:
            self.backend.grub_password = self.grub_manager.grub_password
        else:
            self.backend.grub_password = self.backend.password

    def emit_error(self, text: str) -> None:
        self.lineedit_password.setStyleSheet(self.error_style)
        self.lineedit_password_confirmation.setStyleSheet(self.error_style)
        self.error.emit(self.invalid_password, text)

    def is_equal_passwords(self) -> bool:
        return self.lineedit_password.text() == self.lineedit_password_confirmation.text()

    def show_password_valid(self):
        self.lineedit_password.setStyleSheet("")
        self.lineedit_password_confirmation.setStyleSheet("")
        self.noerror.emit(self.invalid_password)

    def validate_password(self, password: str) -> None:
        valid, error = self.backend.is_valid_password(password)
        if not valid:
            self.emit_error(error)
        elif not self.is_equal_passwords():
            self.emit_error(_('Passwords do not match'))
        else:
            self.show_password_valid()

    def on_lineedit_username_text_changed(self) -> None:
        valid, error = self.backend.is_valid_username(self.lineedit_username.text())
        if valid:
            self.lineedit_username.setStyleSheet("")
            self.noerror.emit(self.invalid_username)
        else:
            self.lineedit_username.setStyleSheet(self.error_style)
            self.error.emit(self.invalid_username, error)

    def on_lineedit_password_text_changed(self) -> None:
        password = self.lineedit_password.text()
        passwordconf = self.lineedit_password_confirmation.text()
        # Терминал и скринлокер (вне киоска)
        # спрашивают пароль, даже если включён автологин
        sysinfo = SettingsManager.systeminfo
        if sysinfo.oem and \
                not sysinfo.is_kiosk_session and \
                len(password) == 0 and \
                len(passwordconf) == 0:
            self.show_password_valid()
        else:
            self.validate_password(password)

    def read_settings(self) -> None:
        self.lineedit_username.setText(self.backend.username)
        self.lineedit_username.setPlaceholderText(self.backend.username)
        self.lineedit_password.setText(self.backend.password)
        self.on_lineedit_password_text_changed()  # TODO: убрать принудительную валидацию
        self.on_lineedit_username_text_changed()
